@@ -31,11 +31,6 @@ def _save_argv(repl=None):
 		_sys.argv[:] = saved
 
 
-@_contextlib.contextmanager
-def null():
-	yield
-
-
 class CustomizedDist(Distribution):
 
 	allow_hosts = None
@@ -98,6 +93,18 @@ class PyTest(orig.test):
 		self.index_url = None
 		self.allow_hosts = None
 		self.addopts = []
+		self.ensure_setuptools_version()
+
+	@staticmethod
+	def ensure_setuptools_version():
+		"""
+		Due to the fact that pytest-runner is often required (via
+		setup-requires directive) by toolchains that never invoke
+		it (i.e. they're only installing the package, not testing it),
+		instead of declaring the dependency in the package
+		metadata, assert the requirement at run time.
+		"""
+		pkg_resources.require('setuptools>=27.3')
 
 	def finalize_options(self):
 		if self.addopts:
@@ -115,21 +122,14 @@ class PyTest(orig.test):
 			and pkg_resources.evaluate_marker(marker)
 		)
 
-	@staticmethod
-	def _install_dists_compat(dist):
-		"""
-		Copy of install_dists from setuptools 27.3.0.
-		"""
-		ir_d = dist.fetch_build_eggs(dist.install_requires or [])
-		tr_d = dist.fetch_build_eggs(dist.tests_require or [])
-		return _itertools.chain(ir_d, tr_d)
-
 	def install_dists(self, dist):
 		"""
 		Extend install_dists to include extras support
 		"""
-		i_d = getattr(orig.test, 'install_dists', self._install_dists_compat)
-		return _itertools.chain(i_d(dist), self.install_extra_dists(dist))
+		return _itertools.chain(
+			orig.test.install_dists(dist),
+			self.install_extra_dists(dist),
+		)
 
 	def install_extra_dists(self, dist):
 		"""
@@ -153,27 +153,18 @@ class PyTest(orig.test):
 		results = list(map(dist.fetch_build_eggs, matching_extras))
 		return _itertools.chain.from_iterable(results)
 
-	@staticmethod
-	def paths_on_pythonpath(paths):
+	def run(self):
 		"""
-		Backward compatibility for paths_on_pythonpath;
-		Returns a null context if paths_on_pythonpath is
-		not implemented in orig.test.
-		Note that this also means that the paths iterable
-		is never consumed, which incidentally means that
-		the None values from dist.fetch_build_eggs in
-		older Setuptools will be disregarded.
+		Override run to ensure requirements are available in this session (but
+		don't install them anywhere).
 		"""
-		try:
-			return orig.test.paths_on_pythonpath(paths)
-		except AttributeError:
-			return null()
-
-	def _super_run(self):
 		dist = CustomizedDist()
 		for attr in 'allow_hosts index_url'.split():
 			setattr(dist, attr, getattr(self, attr))
-		for attr in 'install_requires tests_require extras_require'.split():
+		for attr in (
+			'dependency_links install_requires '
+			'tests_require extras_require '
+		).split():
 			setattr(dist, attr, getattr(self.distribution, attr))
 		installed_dists = self.install_dists(dist)
 		if self.dry_run:
@@ -181,17 +172,8 @@ class PyTest(orig.test):
 			return
 		paths = map(_operator.attrgetter('location'), installed_dists)
 		with self.paths_on_pythonpath(paths):
-			self.with_project_on_sys_path(self.run_tests)
-
-	def run(self):
-		"""
-		Override run to ensure requirements are available in this session (but
-		don't install them anywhere).
-		"""
-		self._super_run()
-		if self.result_code:
-			raise SystemExit(self.result_code)
-		return self.result_code
+			with self.project_on_sys_path():
+				return self.run_tests()
 
 	@property
 	def _argv(self):
@@ -199,7 +181,9 @@ class PyTest(orig.test):
 
 	def run_tests(self):
 		"""
-		Invoke pytest, replacing argv.
+		Invoke pytest, replacing argv. Return result code.
 		"""
 		with _save_argv(_sys.argv[:1] + self.addopts):
-			self.result_code = __import__('pytest').main()
+			result_code = __import__('pytest').main()
+			if result_code:
+				raise SystemExit(result_code)
